            <div class="chatlog__message " data-message-id="{{MESSAGE_ID}}" id="message-{{MESSAGE_ID}}">
                {{MESSAGE_CONTENT}}

                {{EMBEDS}}

                {{EMOJI}}

                {{ATTACHMENTS}}

            </div>